#!/usr/bin/env python3

""" Recursively search and download album covers for a music library. """

import argparse
import asyncio
import base64
import collections
import contextlib
import inspect
import itertools
import logging
import mimetypes
import operator
import os
import string
import sys
import tempfile

import mutagen
import tqdm
import unidecode

import sacad
from sacad import COVER_SOURCE_CLASSES, colored_logging, tqdm_logging

EMBEDDED_ALBUM_ART_SYMBOL = "+"
AUDIO_EXTENSIONS = frozenset(
    ("aac", "ape", "flac", "m4a", "mp3", "mp4", "mpc", "ogg", "oga", "opus", "tta", "wv")
) | frozenset(
    ext for ext, mime in {**mimetypes.types_map, **mimetypes.common_types}.items() if mime.startswith("audio/")
)

Metadata = collections.namedtuple("Metadata", ("artist", "album", "has_embedded_cover"))


# TODO use a dataclasses.dataclass when Python < 3.7 is dropped
class Work:

    """Represent a single search & download work item."""

    def __init__(self, cover_filepath, audio_filepaths, metadata):
        self.cover_filepath = cover_filepath
        self.tmp_cover_filepath = None
        self.audio_filepaths = audio_filepaths
        self.metadata = metadata

    def __repr__(self):
        return (
            f"<{__class__.__qualname__} "
            f"cover_filepath={self.cover_filepath!r} "
            f"tmp_cover_filepath={self.tmp_cover_filepath!r} "
            f"audio_filepaths={self.audio_filepaths!r} "
            f"metadata={self.metadata!r}>"
        )

    def __str__(self):
        return (
            f"cover for {self.metadata.album!r} by {self.metadata.artist!r} "
            f"from {', '.join(map(repr, self.audio_filepaths))}"
        )

    def __eq__(self, other):
        if not isinstance(other, __class__):
            return False
        return (
            (self.cover_filepath == other.cover_filepath)
            and (self.tmp_cover_filepath == other.tmp_cover_filepath)
            and (self.audio_filepaths == other.audio_filepaths)
            and (self.metadata == other.metadata)
        )


def analyze_lib(lib_dir, cover_pattern, *, ignore_existing=False, full_scan=False, all_formats=False):
    """Recursively analyze library, and return a list of work."""
    work = []
    stats = collections.OrderedDict((k, 0) for k in ("files", "albums", "missing covers", "errors"))
    with tqdm.tqdm(desc="Analyzing library", unit="dir", postfix=stats) as progress, tqdm_logging.redirect_logging(
        progress
    ):
        for rootpath, rel_dirpaths, rel_filepaths in os.walk(lib_dir):
            new_work = analyze_dir(
                stats,
                rootpath,
                rel_filepaths,
                cover_pattern,
                ignore_existing=ignore_existing,
                full_scan=full_scan,
                all_formats=all_formats,
            )
            progress.set_postfix(stats, refresh=False)
            progress.update(1)
            work.extend(new_work)
    return work


def get_file_metadata(audio_filepath):
    """Get a Metadata object for this file or None."""
    try:
        mf = mutagen.File(audio_filepath)
    except Exception:
        return
    if mf is None:
        return

    # artist
    for key in ("albumartist", "artist", "TPE1", "TPE2", "aART", "\xa9ART"):  # ogg/ape, mp3, mp4
        try:
            val = mf.get(key, None)
        except ValueError:
            val = None
        if val is not None:
            artist = val[-1]
            break
    else:
        return

    # album
    for key in ("_album", "album", "TALB", "\xa9alb"):  # ogg/ape, mp3, mp4
        try:
            val = mf.get(key, None)
        except ValueError:
            val = None
        if val is not None:
            album = val[-1]
            break
    else:
        return

    # album art
    if isinstance(mf.tags, mutagen._vorbis.VComment):
        has_embedded_cover = "metadata_block_picture" in mf
    elif isinstance(mf.tags, mutagen.id3.ID3):
        has_embedded_cover = any(map(operator.methodcaller("startswith", "APIC:"), mf.keys()))
    elif isinstance(mf.tags, mutagen.mp4.MP4Tags):
        has_embedded_cover = "covr" in mf
    elif isinstance(mf.tags, mutagen.apev2.APEv2):
        has_embedded_cover = "cover art (front)" in mf
    else:
        # unknown tag format
        return

    return Metadata(artist, album, has_embedded_cover)


def get_dir_metadata(audio_filepaths, *, full_scan=False):
    """Build a dict of Metadata to audio filepath list by analyzing audio files."""
    r = collections.defaultdict(list)

    audio_filepaths = tuple(sorted(audio_filepaths))
    for audio_filepath in audio_filepaths:
        file_metadata = get_file_metadata(audio_filepath)
        if file_metadata is None:
            continue

        if not full_scan:
            # stop at the first file that succeeds (for performance)
            # assume all directory files have the same artist/album couple
            r[file_metadata] = audio_filepaths
            break

        r[file_metadata].append(audio_filepath)

    return r


VALID_PATH_CHARS = frozenset(r"-_.()!#$%&'@^{}~" + string.ascii_letters + string.digits)


def sanitize_for_path(s):
    """Sanitize a string to be FAT/NTFS friendly when used in file path."""
    s = s.translate(str.maketrans("/\\|*", "---x"))
    s = "".join(c for c in unidecode.unidecode_expect_ascii(s) if c in VALID_PATH_CHARS)
    s = s.strip()
    s = s.rstrip(".")  # this if for FAT on Android
    return s


def pattern_to_filepath(pattern, parent_dir, metadata):
    """Build absolute cover file path from pattern."""
    assert pattern != EMBEDDED_ALBUM_ART_SYMBOL
    assert metadata.artist is not None
    assert metadata.album is not None
    filepath = pattern.format(artist=sanitize_for_path(metadata.artist), album=sanitize_for_path(metadata.album))
    if not os.path.isabs(filepath):
        filepath = os.path.join(parent_dir, filepath)
    return filepath


def analyze_dir(
    stats, parent_dir, rel_filepaths, cover_pattern, *, ignore_existing=False, full_scan=False, all_formats=False
):
    """Analyze a directory (non recursively) and return a list of Work objects."""
    r = []

    # filter out non audio files
    audio_filepaths = []
    for rel_filepath in rel_filepaths:
        stats["files"] += 1
        try:
            ext = os.path.splitext(rel_filepath)[1][1:].lower()
        except IndexError:
            continue
        if ext in AUDIO_EXTENSIONS:
            audio_filepaths.append(os.path.join(parent_dir, rel_filepath))

    # get metadata
    dir_metadata = get_dir_metadata(audio_filepaths, full_scan=full_scan)

    if audio_filepaths and (not dir_metadata):
        # failed to get any metadata for this directory
        stats["errors"] += 1
        logging.getLogger("sacad_r").error(f"Unable to read metadata for album directory {parent_dir!r}")

    for metadata, album_audio_filepaths in dir_metadata.items():
        # update stats
        stats["albums"] += 1

        # add work item if needed
        if cover_pattern != EMBEDDED_ALBUM_ART_SYMBOL:
            cover_filepath = pattern_to_filepath(cover_pattern, parent_dir, metadata)
            if all_formats:
                missing = ignore_existing or (
                    not any(
                        os.path.isfile(f"{os.path.splitext(cover_filepath)[0]}.{ext}")
                        for ext in sacad.SUPPORTED_IMG_FORMATS
                    )
                )
            else:
                missing = ignore_existing or (not os.path.isfile(cover_filepath))
        else:
            cover_filepath = EMBEDDED_ALBUM_ART_SYMBOL
            missing = (not metadata.has_embedded_cover) or ignore_existing
        if missing:
            stats["missing covers"] += 1
            r.append(Work(cover_filepath, album_audio_filepaths, metadata))

    return r


def embed_album_art(cover_filepath, audio_filepaths):
    """Embed album art into audio files."""
    with open(cover_filepath, "rb") as f:
        cover_data = f.read()

    for filepath in audio_filepaths:
        mf = mutagen.File(filepath)

        if mf.tags is None:
            mf.add_tags()

        if isinstance(mf.tags, mutagen._vorbis.VComment):
            picture = mutagen.flac.Picture()
            picture.data = cover_data
            picture.type = mutagen.id3.PictureType.COVER_FRONT
            picture.mime = "image/jpeg"
            encoded_data = base64.b64encode(picture.write())
            mf["metadata_block_picture"] = encoded_data.decode("ascii")

        elif isinstance(mf.tags, mutagen.id3.ID3):
            mf.tags.add(mutagen.id3.APIC(mime="image/jpeg", type=mutagen.id3.PictureType.COVER_FRONT, data=cover_data))

        elif isinstance(mf.tags, mutagen.mp4.MP4Tags):
            mf["covr"] = [mutagen.mp4.MP4Cover(cover_data, imageformat=mutagen.mp4.AtomDataType.JPEG)]

        elif isinstance(mf.tags, mutagen.apev2.APEv2):
            mf["cover art (front)"] = cover_data

        mf.save()


def ichunk(iterable, n):
    """Split an iterable into n-sized chunks."""
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


def get_covers(work, args):
    """Get missing covers."""
    with contextlib.ExitStack() as cm:

        if args.cover_pattern == EMBEDDED_ALBUM_ART_SYMBOL:
            tmp_prefix = f"{os.path.splitext(os.path.basename(inspect.getfile(inspect.currentframe())))[0]}_"
            tmp_dir = cm.enter_context(tempfile.TemporaryDirectory(prefix=tmp_prefix))

        # setup progress report
        stats = collections.OrderedDict((k, 0) for k in ("ok", "errors", "no result found"))
        progress = cm.enter_context(
            tqdm.tqdm(total=len(work), miniters=1, desc="Searching covers", unit="cover", postfix=stats)
        )
        cm.enter_context(tqdm_logging.redirect_logging(progress))

        def post_download(future):
            work = futures[future]
            try:
                status = future.result()
            except Exception as exception:
                stats["errors"] += 1
                logging.getLogger("sacad_r").error(
                    f"Error occured while searching {work}: {exception.__class__.__qualname__} {exception}"
                )
            else:
                if status:
                    if work.cover_filepath == EMBEDDED_ALBUM_ART_SYMBOL:
                        try:
                            embed_album_art(work.tmp_cover_filepath, work.audio_filepaths)
                        except Exception as exception:
                            stats["errors"] += 1
                            logging.getLogger("sacad_r").error(
                                f"Error occured while embedding {work}: "
                                f"{exception.__class__.__qualname__} {exception}"
                            )
                        else:
                            stats["ok"] += 1
                        finally:
                            os.remove(work.tmp_cover_filepath)
                    else:
                        stats["ok"] += 1
                else:
                    stats["no result found"] += 1
                    logging.getLogger("sacad_r").warning(f"Unable to find {work}")

            progress.set_postfix(stats, refresh=False)
            progress.update(1)

        # post work
        i = 0
        # default event loop on Windows has a 512 fd limit,
        # see https://docs.python.org/3/library/asyncio-eventloops.html#windows
        # also on Linux default max open fd limit is 1024 (ulimit -n)
        # so work in smaller chunks to avoid hitting fd limit
        # this also updates the progress faster (instead of working on all searches, work on finishing the chunk before
        # getting to the next one)
        work_chunk_length = 4 if sys.platform.startswith("win") else 12
        for work_chunk in ichunk(work, work_chunk_length):
            futures = {}
            for i, cur_work in enumerate(work_chunk, i):
                if cur_work.cover_filepath == EMBEDDED_ALBUM_ART_SYMBOL:
                    cover_filepath = os.path.join(tmp_dir, f"{i:02}.{args.format.name.lower()}")
                    cur_work.tmp_cover_filepath = cover_filepath
                else:
                    cover_filepath = cur_work.cover_filepath
                    os.makedirs(os.path.dirname(cover_filepath), exist_ok=True)
                coroutine = sacad.search_and_download(
                    cur_work.metadata.album,
                    cur_work.metadata.artist,
                    args.format,
                    args.size,
                    cover_filepath,
                    size_tolerance_prct=args.size_tolerance_prct,
                    amazon_tlds=args.amazon_tlds,
                    source_classes=args.cover_sources,
                    preserve_format=args.preserve_format,
                )
                future = asyncio.ensure_future(coroutine)
                futures[future] = cur_work

            for future in futures:
                future.add_done_callback(post_download)

            # wait for end of work
            root_future = asyncio.gather(*futures.keys())
            asyncio.get_event_loop().run_until_complete(root_future)


def cl_main():
    """Command line entry point."""
    # parse args
    arg_parser = argparse.ArgumentParser(
        description=f"SACAD (recursive tool) v{sacad.__version__}.{__doc__}",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument("lib_dir", help="Music library directory to recursively analyze")
    arg_parser.add_argument("size", type=int, help="Target image size")
    arg_parser.add_argument(
        "cover_pattern",
        help="""Cover image path pattern.
                {artist} and {album} are replaced by their tag value.
                You can set an absolute path, otherwise destination directory is relative to the audio files.
                Use single character '%s' to embed JPEG into audio files."""
        % (EMBEDDED_ALBUM_ART_SYMBOL),
    )
    arg_parser.add_argument(
        "-i",
        "--ignore-existing",
        action="store_true",
        default=False,
        help="Ignore existing covers and force search and download for all files",
    )
    arg_parser.add_argument(
        "-f",
        "--full-scan",
        action="store_true",
        default=False,
        help="""Enable scanning of all audio files in each directory.
                By default the scanner will assume all audio files in a single directory are part of
                the same album, and only read metadata for the first file.
                Enable this if your files are organized in a way than allows files for different
                albums to be in the same directory level.
                WARNING: This will make the initial scan much slower.""",
    )
    sacad.setup_common_args(arg_parser)
    arg_parser.add_argument(
        "-v", "--verbose", action="store_true", default=False, dest="verbose", help="Enable verbose output"
    )
    args = arg_parser.parse_args()
    if args.cover_pattern == EMBEDDED_ALBUM_ART_SYMBOL:
        args.format = "jpg"
    else:
        args.format = os.path.splitext(args.cover_pattern)[1][1:].lower()
    try:
        args.format = sacad.SUPPORTED_IMG_FORMATS[args.format]
    except KeyError:
        print(f"Unable to guess image format from extension, or unknown format: {args.format}")
        exit(1)
    args.cover_sources = tuple(COVER_SOURCE_CLASSES[source] for source in args.cover_sources)

    # setup logger
    if not args.verbose:
        logging.getLogger("sacad_r").setLevel(logging.WARNING)
        logging.getLogger().setLevel(logging.ERROR)
        logging.getLogger("asyncio").setLevel(logging.CRITICAL + 1)
        fmt = "%(name)s: %(message)s"
    else:
        logging.getLogger("sacad_r").setLevel(logging.DEBUG)
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger("asyncio").setLevel(logging.WARNING)
        fmt = "%(asctime)s %(levelname)s [%(name)s] %(message)s"
    logging_formatter = colored_logging.ColoredFormatter(fmt=fmt)
    logging_handler = logging.StreamHandler()
    logging_handler.setFormatter(logging_formatter)
    logging.getLogger().addHandler(logging_handler)

    # bump nofile ulimit
    try:
        import resource

        soft_lim, hard_lim = resource.getrlimit(resource.RLIMIT_NOFILE)
        if (soft_lim != resource.RLIM_INFINITY) and ((soft_lim < hard_lim) or (hard_lim == resource.RLIM_INFINITY)):
            resource.setrlimit(resource.RLIMIT_NOFILE, (hard_lim, hard_lim))
            logging.getLogger().debug(f"Max open files count set from {soft_lim} to {hard_lim}")
    except (AttributeError, ImportError, OSError, ValueError):
        # not supported on system
        pass

    # do the job
    work = analyze_lib(
        args.lib_dir,
        args.cover_pattern,
        ignore_existing=args.ignore_existing,
        full_scan=args.full_scan,
        all_formats=args.preserve_format,
    )
    get_covers(work, args)


if __name__ == "__main__":
    cl_main()



""" Base class for Amazon cover sources. """

from sacad.sources.base import CoverSource


class AmazonBaseCoverSource(CoverSource):

    """Base class for Amazon cover sources."""

    def __init__(self, *args, base_domain, **kwargs):
        super().__init__(
            *args,
            allow_cookies=True,
            min_delay_between_accesses=2,
            jitter_range_ms=(0, 3000),
            rate_limited_domains=(base_domain,),
            **kwargs,
        )
        self.current_ua = self.ua.firefox
        self.base_domain = base_domain

    def processQueryString(self, s):
        """See CoverSource.processQueryString."""
        return __class__.unaccentuate(__class__.unpunctuate(s.lower()))

    def updateHttpHeaders(self, headers):
        """See CoverSource.updateHttpHeaders."""
        # mimic Firefox headers
        headers["User-Agent"] = self.current_ua
        headers["Accept"] = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"
        headers["Accept-Language"] = "en-US,en;q=0.9"
        headers["DNT"] = "1"
        headers["Connection"] = "Keep-Alive"
        headers["Upgrade-Insecure-Requests"] = "1"
        headers["Cache-Control"] = "max-age=0"
        headers["TE"] = "Trailers"

    def isBlocked(self, html):
        """Return True if Amazon source has blocked our IP (temporarily), and is sending a captcha."""
        blocked_titles = ("Robot Check", "Bot Check", "Amazon CAPTCHA")
        title = html.find("head/title")
        assert title is not None
        return title.text in blocked_titles



""" Amazon CD cover source. """

import collections
import urllib.parse

import lxml.cssselect
import lxml.etree

from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.amazonbase import AmazonBaseCoverSource


class AmazonCdCoverSourceResult(CoverSourceResult):

    """Amazon CD cover search result."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class AmazonCdCoverSource(AmazonBaseCoverSource):

    """Cover source returning Amazon.com audio CD images."""

    TLDS = ("com", "ca", "cn", "fr", "de", "co.jp", "co.uk")
    RESULTS_SELECTORS = (
        lxml.cssselect.CSSSelector("span.rush-component[data-component-type='s-product-image']"),
        lxml.cssselect.CSSSelector("#resultsCol li.s-result-item"),
    )
    IMG_SELECTORS = (lxml.cssselect.CSSSelector("img.s-image"), lxml.cssselect.CSSSelector("img.s-access-image"))
    PRODUCT_LINK_SELECTORS = (lxml.cssselect.CSSSelector("a"), lxml.cssselect.CSSSelector("a.s-access-detail-page"))
    PRODUCT_PAGE_IMG_SELECTOR = lxml.cssselect.CSSSelector("img#landingImage")

    def __init__(self, *args, tld="com", **kwargs):
        assert tld in __class__.TLDS
        self.base_url = f"https://www.amazon.{tld}/s"
        super().__init__(*args, base_domain=urllib.parse.urlsplit(self.base_url).netloc, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        params = collections.OrderedDict()
        params["i"] = "popular"
        params["rh"] = f"p_32:{artist},p_28:{album}"
        params["s"] = "relevancerank"
        return __class__.assembleUrl(self.base_url, params)

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        results = []

        # parse page
        parser = lxml.etree.HTMLParser()
        html = lxml.etree.XML(api_data.decode("utf-8", "ignore"), parser)
        if self.isBlocked(html):
            self.logger.warning("Source is sending a captcha")
            return results

        for page_struct_version, result_selector in enumerate(__class__.RESULTS_SELECTORS):
            result_nodes = result_selector(html)
            if result_nodes:
                break

        for rank, result_node in enumerate(result_nodes, 1):
            try:
                img_node = __class__.IMG_SELECTORS[page_struct_version](result_node)[0]
            except IndexError:
                # no image for that product
                continue
            # get thumbnail & full image url
            thumbnail_url = img_node.get("src")
            url_parts = thumbnail_url.rsplit(".", 2)
            img_url = ".".join((url_parts[0], url_parts[2]))
            # assume size is fixed
            size = (500, 500)
            check_metadata = CoverImageMetadata.SIZE
            # try to get higher res image...
            if (self.target_size > size[0]) and (  # ...only if needed
                rank <= 3
            ):  # and only for first 3 results because this is time
                # consuming (1 more GET request per result)
                product_url = __class__.PRODUCT_LINK_SELECTORS[page_struct_version](result_node)[0].get("href")
                product_url_split = urllib.parse.urlsplit(product_url)
                if not product_url_split.scheme:
                    # relative redirect url
                    product_url_query = urllib.parse.parse_qsl(product_url_split.query)
                    product_url_query = collections.OrderedDict(product_url_query)
                    try:
                        # needed if page_struct_version == 1
                        product_url = product_url_query["url"]
                    except KeyError:
                        # page_struct_version == 0, make url absolute
                        product_url = urllib.parse.urljoin(self.base_url, product_url)
                    product_url_split = urllib.parse.urlsplit(product_url)
                product_url_query = urllib.parse.parse_qsl(product_url_split.query)
                product_url_query = collections.OrderedDict(product_url_query)
                try:
                    # remove timestamp from url to improve future cache hit rate
                    del product_url_query["qid"]
                except KeyError:
                    pass
                product_url_query = urllib.parse.urlencode(product_url_query)
                product_url_no_ts = urllib.parse.urlunsplit(
                    product_url_split[:3] + (product_url_query,) + product_url_split[4:]
                )
                store_in_cache_callback, product_page_data = await self.fetchResults(product_url_no_ts)
                product_page_html = lxml.etree.XML(product_page_data.decode("latin-1"), parser)
                try:
                    img_node = __class__.PRODUCT_PAGE_IMG_SELECTOR(product_page_html)[0]
                except IndexError:
                    # unable to get better image
                    pass
                else:
                    better_img_url = img_node.get("data-old-hires")
                    # img_node.get("data-a-dynamic-image") contains json with image urls too, but they are not larger
                    # than previous 500px image and are often covered by autorip badges (can be removed by cleaning url
                    # though)
                    if better_img_url:
                        img_url = better_img_url
                        size_url_hint = img_url.rsplit(".", 2)[1].strip("_").rsplit("_", 1)[-1]
                        assert size_url_hint.startswith("SL")
                        size_url_hint = int(size_url_hint[2:])
                        size = (size_url_hint, size_url_hint)
                        check_metadata = CoverImageMetadata.NONE
                    await store_in_cache_callback()

            # assume format is always jpg
            format = CoverImageFormat.JPEG
            # add result
            results.append(
                AmazonCdCoverSourceResult(
                    img_url,
                    size,
                    format,
                    thumbnail_url=thumbnail_url,
                    source=self,
                    rank=rank,
                    check_metadata=check_metadata,
                )
            )

        return results



""" Google Images cover source. """

import collections
import json
import urllib.parse

import lxml.cssselect
import lxml.etree

from sacad.cover import SUPPORTED_IMG_FORMATS, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource


class GoogleImagesCoverSourceResult(CoverSourceResult):

    """Google Images cover search result."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, source_quality=CoverSourceQuality.LOW, **kwargs)


class GoogleImagesWebScrapeCoverSource(CoverSource):

    """
    Cover source that scrapes Google Images search result pages.

    Google Image Search JSON API is not used because it is deprecated and Google
    is very agressively rate limiting its access.
    """

    BASE_URL = "https://www.google.com/images"
    RESULTS_SELECTOR = lxml.cssselect.CSSSelector("#search #rg_s .rg_di")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, min_delay_between_accesses=2 / 3, jitter_range_ms=(0, 600), **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        # build request url
        params = collections.OrderedDict()
        params["gbv"] = "2"
        params["q"] = f'"{artist}" "{album}" front cover'
        if abs(self.target_size - 500) < 300:
            params["tbs"] = "isz:m"
        elif self.target_size > 800:
            params["tbs"] = "isz:l"

        return __class__.assembleUrl(__class__.BASE_URL, params)

    def updateHttpHeaders(self, headers):
        """See CoverSource.updateHttpHeaders."""
        headers["User-Agent"] = self.ua.firefox

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        results = []

        # parse HTML and get results
        parser = lxml.etree.HTMLParser()
        html = lxml.etree.XML(api_data.decode("latin-1"), parser)

        for rank, result in enumerate(__class__.RESULTS_SELECTOR(html), 1):
            # extract url
            metadata_div = result.find("div")
            metadata_json = lxml.etree.tostring(metadata_div, encoding="unicode", method="text")
            metadata_json = json.loads(metadata_json)
            google_url = result.find("a").get("href")
            if google_url is not None:
                query = urllib.parse.urlsplit(google_url).query
            else:
                query = None
            if not query:
                img_url = metadata_json["ou"]
            else:
                query = urllib.parse.parse_qs(query)
                img_url = query["imgurl"][0]
            # extract format
            check_metadata = CoverImageMetadata.NONE
            format = metadata_json["ity"].lower()
            try:
                format = SUPPORTED_IMG_FORMATS[format]
            except KeyError:
                # format could not be identified or is unknown
                format = None
                check_metadata = CoverImageMetadata.FORMAT
            # extract size
            if not query:
                size = metadata_json["ow"], metadata_json["oh"]
            else:
                size = tuple(map(int, (query["w"][0], query["h"][0])))
            # extract thumbnail url
            thumbnail_url = metadata_json["tu"]
            # result
            results.append(
                GoogleImagesCoverSourceResult(
                    img_url,
                    size,
                    format,
                    thumbnail_url=thumbnail_url,
                    source=self,
                    rank=rank,
                    check_metadata=check_metadata,
                )
            )

        return results



""" Amazon digital cover source. """

import collections
import operator
import urllib.parse

import lxml.cssselect
import lxml.etree

from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.amazonbase import AmazonBaseCoverSource

AmazonDigitalImageFormat = collections.namedtuple("AmazonDigitalImageFormat", ("id", "slice_count", "total_res"))
AMAZON_DIGITAL_IMAGE_FORMATS = [
    AmazonDigitalImageFormat(
        0, 1, 600
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B00BJ93R7O+c=A17SFUTIVB227Z+d=_SCR(0,0,0)_=.jpg
    AmazonDigitalImageFormat(
        1, 2, 700
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B00BJ93R7O+c=A17SFUTIVB227Z+d=_SCR(1,1,1)_=.jpg
    AmazonDigitalImageFormat(
        1, 4, 1280
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B01NBTSVDN+c=A17SFUTIVB227Z+d=_SCR(1,3,3)_=.jpg
    AmazonDigitalImageFormat(
        2, 3, 1025
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B00BJ93R7O+c=A17SFUTIVB227Z+d=_SCR(2,2,2)_=.jpg
    AmazonDigitalImageFormat(
        2, 5, 1920
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B01NBTSVDN+c=A17SFUTIVB227Z+d=_SCR(2,4,4)_=.jpg
    AmazonDigitalImageFormat(
        3, 4, 1500
    ),  # http://z2-ec2.images-amazon.com/R/1/a=B00BJ93R7O+c=A17SFUTIVB227Z+d=_SCR(3,3,3)_=.jpg
    AmazonDigitalImageFormat(3, 7, 2560),
]  # http://z2-ec2.images-amazon.com/R/1/a=B01NBTSVDN+c=A17SFUTIVB227Z+d=_SCR(3,6,6)_=.jpg
AMAZON_DIGITAL_IMAGE_FORMATS.sort(key=operator.attrgetter("total_res"), reverse=True)


class AmazonDigitalCoverSourceResult(CoverSourceResult):

    """Amazon digital cover search result."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class AmazonDigitalCoverSource(AmazonBaseCoverSource):

    """Cover source returning Amazon.com digital music images."""

    BASE_URL = "https://www.amazon.com"
    DYNAPI_KEY = "A17SFUTIVB227Z"
    RESULTS_SELECTORS = (
        lxml.cssselect.CSSSelector("span.rush-component[data-component-type='s-product-image']"),
        lxml.cssselect.CSSSelector("div#dm_mp3Player li.s-mp3-federated-bar-item"),
    )
    IMG_SELECTORS = (lxml.cssselect.CSSSelector("img.s-image"), lxml.cssselect.CSSSelector("img.s-access-image"))
    LINK_SELECTOR = lxml.cssselect.CSSSelector("a")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, base_domain=urllib.parse.urlsplit(__class__.BASE_URL).netloc, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        url = f"{__class__.BASE_URL}/s"
        params = collections.OrderedDict()
        params["k"] = " ".join((artist, album))
        params["i"] = "digital-music"
        params["s"] = "relevancerank"
        return __class__.assembleUrl(url, params)

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        results = []

        # parse page
        parser = lxml.etree.HTMLParser()
        html = lxml.etree.XML(api_data.decode("utf-8"), parser)
        if self.isBlocked(html):
            self.logger.warning("Source is sending a captcha")
            return results

        for page_struct_version, result_selector in enumerate(__class__.RESULTS_SELECTORS):
            result_nodes = result_selector(html)
            if result_nodes:
                break

        for rank, result_node in enumerate(result_nodes, 1):
            # get thumbnail & full image url
            img_node = __class__.IMG_SELECTORS[page_struct_version](result_node)[0]
            thumbnail_url = img_node.get("src")
            thumbnail_url = thumbnail_url.replace("Stripe-Prime-Only", "")
            url_parts = thumbnail_url.rsplit(".", 2)
            img_url = ".".join((url_parts[0], url_parts[2]))

            # assume size is fixed
            size = (500, 500)

            # try to get higher res image...
            if self.target_size > size[0]:  # ...but only if needed
                self.logger.debug("Looking for optimal subimages configuration...")
                product_url = __class__.LINK_SELECTOR(result_node)[0].get("href")
                product_url = urllib.parse.urlsplit(product_url)
                product_id = product_url.path.split("/")[3]

                # TODO don't pick up highest res image if user asked less?
                for amazon_img_format in AMAZON_DIGITAL_IMAGE_FORMATS:
                    # TODO review this, it seem to always fail now
                    self.logger.debug("Trying %u subimages..." % (amazon_img_format.slice_count ** 2))
                    urls = tuple(
                        self.generateImgUrls(
                            product_id, __class__.DYNAPI_KEY, amazon_img_format.id, amazon_img_format.slice_count
                        )
                    )
                    url_ok = await self.probeUrl(urls[-1])
                    if not url_ok:
                        # images at this size are not available
                        continue

                    # images at this size are available
                    img_url = urls
                    size = (amazon_img_format.total_res,) * 2
                    break

            # assume format is always jpg
            format = CoverImageFormat.JPEG

            # add result
            results.append(
                AmazonDigitalCoverSourceResult(
                    img_url,
                    size,
                    format,
                    thumbnail_url=thumbnail_url,
                    source=self,
                    rank=rank,
                    check_metadata=CoverImageMetadata.SIZE,
                )
            )

        return results

    def generateImgUrls(self, product_id, dynapi_key, format_id, slice_count):
        """Generate URLs for slice_count^2 subimages of a product."""
        for x in range(slice_count):
            for y in range(slice_count):
                yield (
                    "http://z2-ec2.images-amazon.com/R/1/a="
                    + product_id
                    + "+c="
                    + dynapi_key
                    + "+d=_SCR%28"
                    + str(format_id)
                    + ","
                    + str(x)
                    + ","
                    + str(y)
                    + "%29_=.jpg"
                )



""" Deezer cover source. """

import collections
import json
import operator

from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource


class DeezerCoverSourceResult(CoverSourceResult):

    """Deezer search cover result."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, source_quality=CoverSourceQuality.NORMAL, **kwargs)


class DeezerCoverSource(CoverSource):

    """
    Cover source using the official Deezer API.

    https://developers.deezer.com/api/
    """

    BASE_URL = "https://api.deezer.com/search"
    COVER_SIZES = {
        "cover_small": (56, 56),
        "cover": (120, 120),
        "cover_medium": (250, 250),
        "cover_big": (500, 500),
        "cover_xl": (1000, 1000),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, min_delay_between_accesses=0.1, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        # build request url
        search_params = collections.OrderedDict()
        search_params["artist"] = artist
        search_params["album"] = album
        url_params = collections.OrderedDict()
        # url_params["strict"] = "on"
        url_params["order"] = "RANKING"
        url_params["q"] = " ".join(f'{k}:"{v}"' for k, v in search_params.items())
        return __class__.assembleUrl(__class__.BASE_URL, url_params)

    def processQueryString(self, s):
        """See CoverSource.processQueryString."""
        # API search is fuzzy, not need to alter query
        return s

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        results = []

        # get unique albums
        json_data = json.loads(api_data)
        albums = []
        for e in json_data["data"]:
            album = e["album"]
            album_id = album["id"]
            if album_id in map(operator.itemgetter("id"), albums):
                continue
            albums.append(album)

        for rank, album in enumerate(albums, 1):
            for key, size in __class__.COVER_SIZES.items():
                img_url = album[key]
                thumbnail_url = album["cover_small"]
                results.append(
                    DeezerCoverSourceResult(
                        img_url,
                        size,
                        CoverImageFormat.JPEG,
                        thumbnail_url=thumbnail_url,
                        source=self,
                        rank=rank,
                        check_metadata=CoverImageMetadata.NONE,
                    )
                )

        return results



""" Sacad album cover. """

import asyncio
import enum
import imghdr
import io
import itertools
import logging
import math
import mimetypes
import operator
import os
import pickle
import shutil
import urllib.parse

import appdirs
import bitarray

try:
    from bitarray import bitdiff
except ImportError:
    from bitarray.util import count_xor as bitdiff

import PIL.Image
import PIL.ImageFile
import PIL.ImageFilter
import web_cache

from sacad import mkstemp_ctx

PIL.ImageFile.LOAD_TRUNCATED_IMAGES = True

CoverImageFormat = enum.Enum("CoverImageFormat", ("JPEG", "PNG"))

# LOW quality means source can return completely unrelated images
# REFERENCE means the api query is not fuzzy, this is an exact match for the artist/album couple
# NORMAL is in between: reliable source buyt can return wrong album for example
CoverSourceQuality = enum.Enum("CoverSourceQuality", ("LOW", "NORMAL", "REFERENCE"))

CoverImageMetadata = enum.IntEnum("CoverImageMetadata", (("NONE", 0), ("FORMAT", 1), ("SIZE", 2), ("ALL", 3)))

HAS_JPEGOPTIM = shutil.which("jpegoptim") is not None
HAS_OPTIPNG = shutil.which("optipng") is not None
SUPPORTED_IMG_FORMATS = {"jpg": CoverImageFormat.JPEG, "jpeg": CoverImageFormat.JPEG, "png": CoverImageFormat.PNG}
FORMAT_EXTENSIONS = {CoverImageFormat.JPEG: "jpg", CoverImageFormat.PNG: "png"}


def is_square(x):
    """Return True if integer x is a perfect square, False otherwise."""
    return math.sqrt(x).is_integer()


class CoverSourceResult:

    """Cover image returned by a source, candidate to be downloaded."""

    METADATA_PEEK_SIZE_INCREMENT = 2 ** 12
    MAX_FILE_METADATA_PEEK_SIZE = 20 * METADATA_PEEK_SIZE_INCREMENT
    IMG_SIG_SIZE = 16

    def __init__(
        self,
        urls,
        size,
        format,
        *,
        thumbnail_url,
        source,
        source_quality,
        rank=None,
        check_metadata=CoverImageMetadata.NONE,
    ):
        # noqa: D205,D213,D400,D415
        """
        Args:
          urls: Cover image file URL. Can be a tuple of URLs of images to be joined
          size: Cover size as a (with, height) tuple
          format: Cover image format as a CoverImageFormat enum, or None if unknown
          thumbnail_url: Cover thumbnail image file URL, or None if not available
          source: Cover source object that produced this result
          source_quality: Quality of the cover's source as a CoverSourceQuality enum value
          rank: Integer ranking of the cover in the other results from the same source, or None if not available
          check_metadata: If != 0, hint that the format and/or size parameters are not reliable and must be double
          checked
        """
        if not isinstance(urls, str):
            self.urls = urls
        else:
            self.urls = (urls,)
        self.size = size
        assert (format is None) or (format in CoverImageFormat)
        self.format = format
        self.thumbnail_url = thumbnail_url
        self.thumbnail_sig = None
        self.source = source
        assert source_quality in CoverSourceQuality
        self.source_quality = source_quality
        self.rank = rank
        assert (format is not None) or ((check_metadata & CoverImageMetadata.FORMAT) != 0)
        assert (size is not None) or ((check_metadata & CoverImageMetadata.SIZE) != 0)
        self.check_metadata = check_metadata
        self.reliable_metadata = True
        self.is_similar_to_reference = False
        self.is_only_reference = False
        if not hasattr(__class__, "image_cache"):
            cache_filepath = os.path.join(
                appdirs.user_cache_dir(appname="sacad", appauthor=False), "sacad-cache.sqlite"
            )
            os.makedirs(os.path.dirname(cache_filepath), exist_ok=True)
            __class__.image_cache = web_cache.WebCache(
                cache_filepath,
                "cover_image_data",
                caching_strategy=web_cache.CachingStrategy.LRU,
                expiration=60 * 60 * 24 * 365,
            )  # 1 year
            __class__.metadata_cache = web_cache.WebCache(
                cache_filepath,
                "cover_metadata",
                caching_strategy=web_cache.CachingStrategy.LRU,
                expiration=60 * 60 * 24 * 365,
            )  # 1 year
            for cache, cache_name in zip(
                (__class__.image_cache, __class__.metadata_cache), ("cover_image_data", "cover_metadata")
            ):
                purged_count = cache.purge()
                logging.getLogger("Cache").debug(
                    f"{purged_count} obsolete entries have been removed from cache {cache_name!r}"
                )
                row_count = len(cache)
                logging.getLogger("Cache").debug(f"Cache {cache_name!r} contains {row_count} entries")

    def __str__(self):
        s = f"{self.__class__.__name__} {self.urls[0]!r}"
        if len(self.urls) > 1:
            s += f" [x{len(self.urls)}]"
        return s

    async def get(self, target_format, target_size, size_tolerance_prct, out_filepath, *, preserve_format=False):
        """Download cover and process it."""
        if self.source_quality.value <= CoverSourceQuality.LOW.value:
            logging.getLogger("Cover").warning(
                "Cover is from a potentially unreliable source and may be unrelated to the search"
            )

        images_data = []
        for i, url in enumerate(self.urls):
            # download
            logging.getLogger("Cover").info(f"Downloading cover {url!r} (part {i + 1}/{len(self.urls)})...")
            headers = {}
            self.source.updateHttpHeaders(headers)

            async def pre_cache_callback(img_data):
                return await __class__.crunch(img_data, self.format)

            store_in_cache_callback, image_data = await self.source.http.query(
                url, headers=headers, verify=False, cache=__class__.image_cache, pre_cache_callback=pre_cache_callback
            )

            # store immediately in cache
            await store_in_cache_callback()

            # append for multi images
            images_data.append(image_data)

        need_format_change = self.format != target_format
        need_size_change = (max(self.size) > target_size) and (
            abs(max(self.size) - target_size) > target_size * size_tolerance_prct / 100
        )
        need_join = len(images_data) > 1
        if (need_format_change and (not preserve_format)) or need_join or need_size_change:
            # post process
            image_data = self.postProcess(
                images_data, target_format if need_format_change else None, target_size if need_size_change else None
            )

            # crunch image again
            image_data = await __class__.crunch(image_data, target_format)

            format_changed = need_format_change
        else:
            format_changed = False

        # write it
        if need_format_change and (not format_changed):
            assert preserve_format
            out_filepath = f"{os.path.splitext(out_filepath)[0]}.{FORMAT_EXTENSIONS[self.format]}"
        with open(out_filepath, "wb") as file:
            file.write(image_data)

    def postProcess(self, images_data, new_format, new_size):
        """
        Convert image binary data.

        Convert image binary data to a target format and/or size (None if no conversion needed), and return the
        processed data.
        """
        if len(images_data) == 1:
            in_bytes = io.BytesIO(images_data[0])
            img = PIL.Image.open(in_bytes)
            if img.mode != "RGB":
                img = img.convert("RGB")

        else:
            # images need to be joined before further processing
            logging.getLogger("Cover").info(f"Joining {len(images_data)} images...")
            # TODO find a way to do this losslessly for JPEG
            new_img = PIL.Image.new("RGB", self.size)
            assert is_square(len(images_data))
            sq = int(math.sqrt(len(images_data)))

            images_data_it = iter(images_data)
            img_sizes = {}
            for x in range(sq):
                for y in range(sq):
                    current_image_data = next(images_data_it)
                    img_stream = io.BytesIO(current_image_data)
                    img = PIL.Image.open(img_stream)
                    img_sizes[(x, y)] = img.size
                    box = [0, 0]
                    if x > 0:
                        for px in range(x):
                            box[0] += img_sizes[(px, y)][0]
                    if y > 0:
                        for py in range(y):
                            box[1] += img_sizes[(x, py)][1]
                    box.extend((box[0] + img.size[0], box[1] + img.size[1]))
                    new_img.paste(img, box=tuple(box))
            img = new_img

        out_bytes = io.BytesIO()
        if new_size is not None:
            logging.getLogger("Cover").info(f"Resizing from {self.size[0]}x{self.size[1]} to {new_size}x{new_size}...")
            img = img.resize((new_size, new_size), PIL.Image.LANCZOS)
            # apply unsharp filter to remove resize blur (equivalent to (images/graphics)magick -unsharp 1.5x1+0.7+0.02)
            # we don't use PIL.ImageFilter.SHARPEN or PIL.ImageEnhance.Sharpness because we want precise control over
            # parameters
            unsharper = PIL.ImageFilter.UnsharpMask(radius=1.5, percent=70, threshold=5)
            img = img.filter(unsharper)
        if new_format is not None:
            logging.getLogger("Cover").info(f"Converting to {new_format.name.upper()}...")
            target_format = new_format
        else:
            target_format = self.format
        img.save(out_bytes, format=target_format.name, quality=90, optimize=True)
        return out_bytes.getvalue()

    async def updateImageMetadata(self):  # noqa: C901
        """Download image file(s) partially to get its real metadata, or get it from cache."""
        assert self.needMetadataUpdate()

        width_sum, height_sum = 0, 0

        # only download metadata for the needed images to get full size
        idxs = []
        assert is_square(len(self.urls))
        sq = int(math.sqrt(len(self.urls)))
        for x in range(sq):
            for y in range(sq):
                if x == y:
                    idxs.append((x * sq + y, x, y))

        for idx, x, y in idxs:
            url = self.urls[idx]
            format, width, height = None, None, None

            try:
                format, width, height = pickle.loads(__class__.metadata_cache[url])
            except KeyError:
                # cache miss
                pass
            except Exception as e:
                logging.getLogger("Cover").warning(
                    f"Unable to load metadata for URL {url!r} from cache: {e.__class__.__qualname__} {e}"
                )
            else:
                # cache hit
                logging.getLogger("Cover").debug(f"Got metadata for URL {url!r} from cache")
                if format is not None:
                    self.setFormatMetadata(format)

            if self.needMetadataUpdate(CoverImageMetadata.FORMAT) or (
                self.needMetadataUpdate(CoverImageMetadata.SIZE) and ((width is None) or (height is None))
            ):
                # download
                logging.getLogger("Cover").debug(f"Downloading file header for URL {url!r}...")
                try:
                    headers = {}
                    self.source.updateHttpHeaders(headers)
                    response = await self.source.http.fastStreamedQuery(url, headers=headers, verify=False)
                    try:
                        if self.needMetadataUpdate(CoverImageMetadata.FORMAT):
                            # try to get format from response
                            format = __class__.guessImageFormatFromHttpResponse(response)
                            if format is not None:
                                self.setFormatMetadata(format)

                        if self.needMetadataUpdate():
                            # try to get metadata from HTTP data
                            metadata = await __class__.guessImageMetadataFromHttpData(response)
                            if metadata is not None:
                                format, width, height = metadata
                                if format is not None:
                                    self.setFormatMetadata(format)

                    finally:
                        await response.release()

                except Exception as e:
                    logging.getLogger("Cover").warning(
                        f"Failed to get file metadata for URL {url!r} ({e.__class__.__qualname__} {e})"
                    )

                if self.needMetadataUpdate():  # did we fail to get needed metadata at this point?
                    if (self.format is None) or ((self.size is None) and ((width is None) or (height is None))):
                        # if we get here, file is probably not reachable, or not even an image
                        logging.getLogger("Cover").debug(
                            f"Unable to get file metadata from file or HTTP headers for URL {url!r}, "
                            "skipping this result"
                        )
                        return

                    if (self.format is not None) and ((self.size is not None) and (width is None) and (height is None)):
                        logging.getLogger("Cover").debug(
                            f"Unable to get file metadata from file or HTTP headers for URL {url!r}, "
                            "falling back to API data" % (url)
                        )
                        self.check_metadata = CoverImageMetadata.NONE
                        self.reliable_metadata = False
                        return

                # save it to cache
                __class__.metadata_cache[url] = pickle.dumps((format, width, height))

            # sum sizes
            if (width is not None) and (height is not None):
                width_sum += width
                height_sum += height

        if self.needMetadataUpdate(CoverImageMetadata.SIZE) and (width_sum > 0) and (height_sum > 0):
            self.setSizeMetadata((width_sum, height_sum))

    def needMetadataUpdate(self, what=CoverImageMetadata.ALL):
        """Return True if image metadata needs to be checked, False instead."""
        return (self.check_metadata & what) != 0

    def setFormatMetadata(self, format):
        """Set format image metadata to what has been reliably identified."""
        self.format = format
        self.check_metadata &= ~CoverImageMetadata.FORMAT

    def setSizeMetadata(self, size):
        """Set size image metadata to what has been reliably identified."""
        self.size = size
        self.check_metadata &= ~CoverImageMetadata.SIZE

    async def updateSignature(self):
        """Calculate a cover's "signature" using its thumbnail url."""
        assert self.thumbnail_sig is None

        if self.thumbnail_url is None:
            logging.getLogger("Cover").warning(f"No thumbnail available for {self}")
            return

        # download
        logging.getLogger("Cover").debug(f"Downloading cover thumbnail {self.thumbnail_url!r}...")
        headers = {}
        self.source.updateHttpHeaders(headers)

        async def pre_cache_callback(img_data):
            return await __class__.crunch(img_data, CoverImageFormat.JPEG, silent=True)

        try:
            store_in_cache_callback, image_data = await self.source.http.query(
                self.thumbnail_url, cache=__class__.image_cache, headers=headers, pre_cache_callback=pre_cache_callback
            )
        except Exception as e:
            logging.getLogger("Cover").warning(
                f"Download of {self.thumbnail_url!r} failed: {e.__class__.__qualname__} {e}"
            )
            return

        # compute sig
        logging.getLogger("Cover").debug(f"Computing signature of {self}...")
        try:
            self.thumbnail_sig = __class__.computeImgSignature(image_data)
        except Exception as e:
            logging.getLogger("Cover").warning(
                f"Failed to compute signature of '{self}': {e.__class__.__qualname__} {e}"
            )
        else:
            await store_in_cache_callback()

    @staticmethod
    def compare(first, second, *, target_size, size_tolerance_prct):
        """
        Compare cover relevance/quality.

        Return -1 if first is a worst match than second, 1 otherwise, or 0 if cover can't be discriminated.

        This code is responsible for comparing two cover results to identify the best one, and is used to sort all
        results.
        It is probably the most important piece of code of this tool.
        Covers with sizes under the target size (+- configured tolerance) are excluded before comparison.
        The following factors are used in order:
          1. Prefer approximately square covers
          2. Prefer covers similar to the reference cover
          3. Prefer size above target size
          4. If both below target size, prefer closest
          5. Prefer covers of most reliable source
          6. Prefer best ranked cover
          7. Prefer covers with reliable metadata
        If all previous factors do not allow sorting of two results (very unlikely):
          8. Prefer covers with less images to join
          9. Prefer covers having the target size
          10. Prefer PNG covers
          11. Prefer exactly square covers

        We don't overload the __lt__ operator because we need to pass the target_size parameter.

        """
        for c in (first, second):
            assert c.format is not None
            assert isinstance(c.size[0], int) and isinstance(c.size[1], int)

        # prefer square covers #1
        delta_ratio1 = abs(first.size[0] / first.size[1] - 1)
        delta_ratio2 = abs(second.size[0] / second.size[1] - 1)
        if abs(delta_ratio1 - delta_ratio2) > 0.15:
            return -1 if (delta_ratio1 > delta_ratio2) else 1

        # prefer similar to reference
        sr1 = first.is_similar_to_reference
        sr2 = second.is_similar_to_reference
        if sr1 and (not sr2):
            return 1
        if (not sr1) and sr2:
            return -1

        # prefer size above preferred
        delta_size1 = ((first.size[0] + first.size[1]) / 2) - target_size
        delta_size2 = ((second.size[0] + second.size[1]) / 2) - target_size
        if ((delta_size1 < 0) and (delta_size2 >= 0)) or (delta_size1 >= 0) and (delta_size2 < 0):
            return -1 if (delta_size1 < delta_size2) else 1

        # if both below target size, prefer closest
        if (delta_size1 < 0) and (delta_size2 < 0) and (delta_size1 != delta_size2):
            return -1 if (delta_size1 < delta_size2) else 1

        # prefer covers of most reliable source
        qs1 = first.source_quality.value
        qs2 = second.source_quality.value
        if qs1 != qs2:
            return -1 if (qs1 < qs2) else 1

        # prefer best ranked
        if (
            (first.rank is not None)
            and (second.rank is not None)
            and (first.__class__ is second.__class__)
            and (first.rank != second.rank)
        ):
            return -1 if (first.rank > second.rank) else 1

        # prefer reliable metadata
        if first.reliable_metadata != second.reliable_metadata:
            return 1 if first.reliable_metadata else -1

        # prefer covers with less images to join
        ic1 = len(first.urls)
        ic2 = len(second.urls)
        if ic1 != ic2:
            return -1 if (ic1 > ic2) else 1

        # prefer the preferred size
        if abs(delta_size1) != abs(delta_size2):
            return -1 if (abs(delta_size1) > abs(delta_size2)) else 1

        # prefer png
        if first.format != second.format:
            return -1 if (second.format is CoverImageFormat.PNG) else 1

        # prefer square covers #2
        if delta_ratio1 != delta_ratio2:
            return -1 if (delta_ratio1 > delta_ratio2) else 1

        # fuck, they are the same!
        return 0

    @staticmethod
    async def crunch(image_data, format, silent=False):
        """Crunch image data, and return the processed data, or orignal data if operation failed."""
        if ((format is CoverImageFormat.PNG) and (not HAS_OPTIPNG)) or (
            (format is CoverImageFormat.JPEG) and (not HAS_JPEGOPTIM)
        ):
            return image_data
        with mkstemp_ctx.mkstemp(suffix=f".{format.name.lower()}") as tmp_out_filepath:
            if not silent:
                logging.getLogger("Cover").info(f"Crunching {format.name.upper()} image...")
            with open(tmp_out_filepath, "wb") as tmp_out_file:
                tmp_out_file.write(image_data)
            size_before = len(image_data)
            if format is CoverImageFormat.PNG:
                cmd = ["optipng", "-quiet", "-o1"]
            elif format is CoverImageFormat.JPEG:
                cmd = ["jpegoptim", "-q", "--strip-all"]
            cmd.append(tmp_out_filepath)
            p = await asyncio.create_subprocess_exec(
                *cmd,
                stdin=asyncio.subprocess.DEVNULL,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
            )
            await p.wait()
            if p.returncode != 0:
                if not silent:
                    logging.getLogger("Cover").warning("Crunching image failed")
                return image_data
            with open(tmp_out_filepath, "rb") as tmp_out_file:
                crunched_image_data = tmp_out_file.read()
            size_after = len(crunched_image_data)
            pct_saved = 100 * (size_before - size_after) / size_before
            if not silent:
                logging.getLogger("Cover").debug(f"Crunching image saved {pct_saved:.2f}% filesize")
        return crunched_image_data

    @staticmethod
    def guessImageMetadataFromData(img_data):
        """Identify an image format and size from its first bytes."""
        format, width, height = None, None, None
        img_stream = io.BytesIO(img_data)
        try:
            img = PIL.Image.open(img_stream)
        except (IOError, RuntimeError):
            format = imghdr.what(None, h=img_data)
            format = SUPPORTED_IMG_FORMATS.get(format, None)
        else:
            format = img.format.lower()
            format = SUPPORTED_IMG_FORMATS.get(format, None)
            width, height = img.size
        return format, width, height

    @staticmethod
    async def guessImageMetadataFromHttpData(response):
        """Identify an image format and size from the beginning of its HTTP data."""
        metadata = None
        img_data = bytearray()

        while len(img_data) < CoverSourceResult.MAX_FILE_METADATA_PEEK_SIZE:
            new_img_data = await response.content.read(__class__.METADATA_PEEK_SIZE_INCREMENT)
            if not new_img_data:
                break
            img_data.extend(new_img_data)

            metadata = __class__.guessImageMetadataFromData(img_data)
            if (metadata is not None) and all(metadata):
                return metadata

        return metadata

    @staticmethod
    def guessImageFormatFromHttpResponse(response):
        """Guess file format from HTTP response, return format or None."""
        extensions = []

        # try to guess extension from response content-type header
        try:
            content_type = response.headers["Content-Type"]
        except KeyError:
            pass
        else:
            ext = mimetypes.guess_extension(content_type, strict=False)
            if ext is not None:
                extensions.append(ext)

        # try to extract extension from URL
        urls = list(response.history) + [response.url]
        for url in map(str, urls):
            ext = os.path.splitext(urllib.parse.urlsplit(url).path)[-1]
            if (ext is not None) and (ext not in extensions):
                extensions.append(ext)

        # now guess from the extensions
        for ext in extensions:
            try:
                return SUPPORTED_IMG_FORMATS[ext[1:]]
            except KeyError:
                pass

    @staticmethod
    async def preProcessForComparison(results, target_size, size_tolerance_prct):
        """Process results to prepare them for future comparison and sorting."""
        # find reference (=image most likely to match target cover ignoring factors like size and format)
        reference = None
        for result in results:
            if result.source_quality is CoverSourceQuality.REFERENCE:
                if (reference is None) or (
                    CoverSourceResult.compare(
                        result, reference, target_size=target_size, size_tolerance_prct=size_tolerance_prct
                    )
                    > 0
                ):
                    reference = result

        # remove results that are only refs
        results = list(itertools.filterfalse(operator.attrgetter("is_only_reference"), results))

        # remove duplicates
        no_dup_results = []
        for result in results:
            is_dup = False
            for result_comp in results:
                if (
                    (result_comp is not result)
                    and (result_comp.urls == result.urls)
                    and (
                        __class__.compare(
                            result, result_comp, target_size=target_size, size_tolerance_prct=size_tolerance_prct
                        )
                        < 0
                    )
                ):
                    is_dup = True
                    break
            if not is_dup:
                no_dup_results.append(result)
        dup_count = len(results) - len(no_dup_results)
        if dup_count > 0:
            logging.getLogger("Cover").info(f"Removed {dup_count} duplicate results")
            results = no_dup_results

        if reference is not None:
            logging.getLogger("Cover").info(f"Reference is: {reference}")
            reference.is_similar_to_reference = True

            # calculate sigs
            futures = []
            for result in results:
                coroutine = result.updateSignature()
                future = asyncio.ensure_future(coroutine)
                futures.append(future)
            if reference.is_only_reference:
                assert reference not in results
                coroutine = reference.updateSignature()
                future = asyncio.ensure_future(coroutine)
                futures.append(future)
            if futures:
                await asyncio.wait(futures)
            for future in futures:
                future.result()  # raise pending exception if any

            # compare other results to reference
            for result in results:
                if (
                    (result is not reference)
                    and (result.thumbnail_sig is not None)
                    and (reference.thumbnail_sig is not None)
                ):
                    result.is_similar_to_reference = __class__.areImageSigsSimilar(
                        result.thumbnail_sig, reference.thumbnail_sig
                    )
                    if result.is_similar_to_reference:
                        logging.getLogger("Cover").debug(f"{result} is similar to reference")
                    else:
                        logging.getLogger("Cover").debug(f"{result} is NOT similar to reference")
        else:
            logging.getLogger("Cover").warning("No reference result found")

        return results

    @staticmethod
    def computeImgSignature(image_data):
        """
        Calculate an image signature.

        This is similar to ahash but uses 3 colors components
        See: https://github.com/JohannesBuchner/imagehash/blob/4.0/imagehash/__init__.py#L125

        """
        parser = PIL.ImageFile.Parser()
        parser.feed(image_data)
        img = parser.close()
        target_size = (__class__.IMG_SIG_SIZE, __class__.IMG_SIG_SIZE)
        img.thumbnail(target_size, PIL.Image.BICUBIC)
        if img.size != target_size:
            logging.getLogger("Cover").debug(
                "Non square thumbnail after resize to %ux%u, unable to compute signature" % target_size
            )
            return None
        img = img.convert(mode="RGB")
        pixels = img.getdata()
        pixel_count = target_size[0] * target_size[1]
        color_count = 3
        r = bitarray.bitarray(pixel_count * color_count)
        r.setall(False)
        for ic in range(color_count):
            mean = sum(p[ic] for p in pixels) // pixel_count
            for ip, p in enumerate(pixels):
                if p[ic] > mean:
                    r[pixel_count * ic + ip] = True
        return r

    @staticmethod
    def areImageSigsSimilar(sig1, sig2):
        """Compare 2 image "signatures" and return True if they seem to come from a similar image, False otherwise."""
        return bitdiff(sig1, sig2) < 100


# silence third party module loggers
logging.getLogger("PIL").setLevel(logging.ERROR)



""" SACAD cover sources. """

from sacad.sources.amazoncd import AmazonCdCoverSource, AmazonCdCoverSourceResult  # noqa: F401
from sacad.sources.amazondigital import AmazonDigitalCoverSource, AmazonDigitalCoverSourceResult  # noqa: F401
from sacad.sources.deezer import DeezerCoverSource, DeezerCoverSourceResult  # noqa: F401
from sacad.sources.google_images import GoogleImagesCoverSourceResult, GoogleImagesWebScrapeCoverSource  # noqa: F401
from sacad.sources.lastfm import LastFmCoverSource, LastFmCoverSourceResult  # noqa: F401



""" Code to help using the logging module with tqdm progress bars. """

import contextlib
import logging
import threading

logging_handlers_lock = threading.Lock()


class TqdmLoggingHandler(logging.Handler):

    """Logging handler sending messages to the tqdm write method (avoids overlap)."""

    def __init__(self, tqdm, *args, **kwargs):
        self.tqdm = tqdm
        super().__init__(*args, **kwargs)

    def emit(self, record):
        """See logging.Handler.emit."""
        msg = self.format(record)
        self.tqdm.write(msg)


@contextlib.contextmanager
def redirect_logging(tqdm_obj, logger=logging.getLogger()):
    """Redirect logging to a TqdmLoggingHandler object and then restore the original logging behavior."""
    with logging_handlers_lock:
        # remove current handlers
        prev_handlers = []
        for handler in logger.handlers.copy():
            prev_handlers.append(handler)
            logger.removeHandler(handler)

        # add tqdm handler
        tqdm_handler = TqdmLoggingHandler(tqdm_obj)
        if prev_handlers[-1].formatter is not None:
            tqdm_handler.setFormatter(prev_handlers[-1].formatter)
        logger.addHandler(tqdm_handler)

    try:
        yield
    finally:
        # restore handlers
        with logging_handlers_lock:
            logger.removeHandler(tqdm_handler)
            for handler in prev_handlers:
                logger.addHandler(handler)



#!/usr/bin/env python3

""" Package setup. """

import os
import re
import sys

from setuptools import find_packages, setup

if sys.hexversion < 0x3060000:
    print("Python version %s is unsupported, >= 3.6.0 is needed" % (".".join(map(str, sys.version_info[:3]))))
    exit(1)

with open(os.path.join("sacad", "__init__.py"), "rt") as f:
    version_match = re.search('__version__ = "([^"]+)"', f.read())
assert version_match is not None
version = version_match.group(1)

with open("requirements.txt", "rt") as f:
    requirements = f.read().splitlines()
if sys.platform == "win32":
    requirements.append("idna==2.6")

with open("test-requirements.txt", "rt") as f:
    test_requirements = f.read().splitlines()

with open("README.md", "rt") as f:
    readme = f.read()

setup(
    name="sacad",
    version=version,
    author="desbma",
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": ["sacad = sacad:cl_main", "sacad_r = sacad.recurse:cl_main"]},
    test_suite="tests",
    install_requires=requirements,
    tests_require=test_requirements,
    description="Search and download music album covers",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/desbma/sacad",
    download_url="https://github.com/desbma/sacad/archive/%s.tar.gz" % (version),
    keywords=["download", "album", "cover", "art", "albumart", "music"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Utilities",
    ],
)
